import { Buffer } from 'buffer';
export var DataTypes;
(function (DataTypes) {
    DataTypes[DataTypes["BYTE"] = 0] = "BYTE";
    DataTypes[DataTypes["BOOL"] = 1] = "BOOL";
    DataTypes[DataTypes["INT8"] = 2] = "INT8";
    DataTypes[DataTypes["INT16"] = 3] = "INT16";
    DataTypes[DataTypes["INT32"] = 4] = "INT32";
    DataTypes[DataTypes["UINT8"] = 5] = "UINT8";
    DataTypes[DataTypes["UINT16"] = 6] = "UINT16";
    DataTypes[DataTypes["UINT32"] = 7] = "UINT32";
    DataTypes[DataTypes["FLOAT"] = 8] = "FLOAT";
    DataTypes[DataTypes["DOUBLE"] = 9] = "DOUBLE";
    DataTypes[DataTypes["DATE"] = 10] = "DATE";
    DataTypes[DataTypes["STRING"] = 11] = "STRING";
    DataTypes[DataTypes["BYTE_ARRAY"] = 12] = "BYTE_ARRAY";
})(DataTypes || (DataTypes = {}));
export class ByteList {
    constructor(bytes) {
        this.useLittleEndian = true;
        this._paddingSize = 100;
        this.index = 0;
        this._length = 0;
        this._buffer = Buffer.alloc(this._paddingSize);
        if (typeof bytes === 'string' || bytes instanceof String) {
            this.writeString(bytes);
        }
        else if (bytes) {
            this.concat(bytes);
        }
        this.index = 0;
    }
    static SetBit(num, val, bit) {
        if (val) {
            num |= 1 << bit;
        }
        else {
            num &= ~(1 << bit);
        }
        return num;
    }
    static GetBit(num, bit) {
        return (num & (1 << bit)) != 0;
    }
    get length() { return this._length; }
    get buffer() { return this.getBuffer(); }
    get paddingSize() { return this._paddingSize; }
    set paddingSize(val) {
        this._paddingSize = val;
        if (this._paddingSize < 0) {
            this._paddingSize = 0;
        }
    }
    getBuffer() {
        return Buffer.from(this._buffer.buffer, 0, this._length);
    }
    getLength() {
        return this._length;
    }
    concat(buffer) {
        if (buffer instanceof ArrayBuffer) {
            buffer = new Uint8Array(buffer);
        }
        else if (buffer && buffer.buffer && buffer.buffer instanceof Uint8Array) {
            buffer = buffer.buffer;
        }
        this.prepareBuffer(buffer.length);
        for (let i = 0; i < buffer.length; i++) {
            if (this.index === this._length) {
                this.index++;
            }
            this._buffer.writeUInt8(buffer[i], this._length++);
        }
    }
    insert(buffer) {
        if (buffer instanceof ByteList) {
            buffer = buffer._buffer;
        }
        const part1 = this._buffer.slice(0, this.index);
        const part2 = this._buffer.slice(this.index, this._buffer.length);
        this._buffer = Buffer.concat([part1, buffer, part2], (part1.length + buffer.length + part2.length));
        this.index += buffer.length;
        this._length += buffer.length;
    }
    peekByte(offset = 0) {
        if (this.length < this.index + offset + 1) {
            throw new Error('Buffer Overrun');
        }
        return this._buffer.readUInt8(this.index + offset);
    }
    peekUInt16(offset = 0) {
        if (this.length < this.index + offset + 2) {
            throw new Error('Buffer Overrun');
        }
        if (this.useLittleEndian) {
            return this._buffer.readUInt16LE(this.index + offset);
        }
        else {
            return this._buffer.readUInt16BE(this.index + offset);
        }
    }
    writeByte(byte, options = {}) {
        if (typeof byte === 'string' && byte.length === 1) {
            byte = byte.charCodeAt(0);
        }
        this.prepareBuffer(1);
        const buf = options.insert ? Buffer.alloc(1) : this._buffer;
        buf.writeUInt8(!byte || byte < 0 ? 0 : byte, options.insert ? 0 : this.index);
        if (options.insert) {
            this.insert(buf);
        }
        else {
            this.index += 1;
            this._length += 1;
        }
    }
    writeBool(bool, options = {}) {
        this.writeByte(bool ? 1 : 0, options);
    }
    writeInt16(int16, options = {}) {
        this.prepareBuffer(2);
        const buf = options.insert ? Buffer.alloc(2) : this._buffer;
        if (this.useLittleEndian) {
            buf.writeInt16LE(!int16 ? 0 : int16, options.insert ? 0 : this.index);
        }
        else {
            buf.writeInt16BE(!int16 ? 0 : int16, options.insert ? 0 : this.index);
        }
        if (options.insert) {
            this.insert(buf);
        }
        else {
            this.index += 2;
            this._length += 2;
        }
    }
    writeInt32(int32, options = {}) {
        this.prepareBuffer(4);
        const buf = options.insert ? Buffer.alloc(4) : this._buffer;
        if (this.useLittleEndian) {
            buf.writeInt32LE(!int32 ? 0 : int32, options.insert ? 0 : this.index);
        }
        else {
            buf.writeInt32BE(!int32 ? 0 : int32, options.insert ? 0 : this.index);
        }
        if (options.insert) {
            this.insert(buf);
        }
        else {
            this.index += 4;
            this._length += 4;
        }
    }
    writeUInt16(uint16, options = {}) {
        this.prepareBuffer(2);
        const buf = options.insert ? Buffer.alloc(2) : this._buffer;
        if (this.useLittleEndian) {
            buf.writeUInt16LE(!uint16 ? 0 : uint16, options.insert ? 0 : this.index);
        }
        else {
            buf.writeUInt16BE(!uint16 ? 0 : uint16, options.insert ? 0 : this.index);
        }
        if (options.insert) {
            this.insert(buf);
        }
        else {
            this.index += 2;
            this._length += 2;
        }
    }
    writeUInt32(uint32, options = {}) {
        this.prepareBuffer(4);
        const buf = options.insert ? Buffer.alloc(4) : this._buffer;
        if (this.useLittleEndian) {
            buf.writeUInt32LE(!uint32 ? 0 : uint32, options.insert ? 0 : this.index);
        }
        else {
            buf.writeUInt32BE(!uint32 ? 0 : uint32, options.insert ? 0 : this.index);
        }
        if (options.insert) {
            this.insert(buf);
        }
        else {
            this.index += 4;
            this._length += 4;
        }
    }
    writeFloat(float, options = {}) {
        this.prepareBuffer(8);
        const buf = options.insert ? Buffer.alloc(8) : this._buffer;
        if (this.useLittleEndian) {
            buf.writeFloatLE(!float ? 0 : float, options.insert ? 0 : this.index);
        }
        else {
            buf.writeFloatBE(!float ? 0 : float, options.insert ? 0 : this.index);
        }
        if (options.insert) {
            this.insert(buf);
        }
        else {
            this.index += 8;
            this._length += 8;
        }
    }
    writeDouble(double, options = {}) {
        this.prepareBuffer(8);
        const buf = options.insert ? Buffer.alloc(8) : this._buffer;
        if (this.useLittleEndian) {
            buf.writeDoubleLE(!double ? 0 : double, options.insert ? 0 : this.index);
        }
        else {
            buf.writeDoubleBE(!double ? 0 : double, options.insert ? 0 : this.index);
        }
        if (options.insert) {
            this.insert(buf);
        }
        else {
            this.index += 8;
            this._length += 8;
        }
    }
    writeDate(date, options = {}) {
        this.prepareBuffer(6);
        const buffer = options.insert ? Buffer.alloc(6) : this._buffer;
        buffer.writeUInt8(!date ? 0 : date.getUTCFullYear() - 2000, options.insert ? 0 : this.index++);
        buffer.writeUInt8(!date ? 0 : date.getUTCMonth(), options.insert ? 1 : this.index++);
        buffer.writeUInt8(!date ? 0 : date.getUTCDate(), options.insert ? 2 : this.index++);
        buffer.writeUInt8(!date ? 0 : date.getUTCHours(), options.insert ? 3 : this.index++);
        buffer.writeUInt8(!date ? 0 : date.getUTCMinutes(), options.insert ? 4 : this.index++);
        buffer.writeUInt8(!date ? 0 : date.getUTCSeconds(), options.insert ? 5 : this.index++);
        if (options.insert) {
            this.insert(buffer);
        }
        else {
            this._length += 6;
        }
    }
    writeString(str = '', options = {}) {
        if (!options.length) {
            const buf = Buffer.from(str, 'ascii');
            this.writeUInt16(buf.length, options);
            this.concat(buf);
        }
        else {
            const buf = Buffer.from(str.length > options.length ? str.substr(0, options.length) : str, 'ascii');
            console.log(buf);
            this.concat(buf);
            for (let i = 0; i < options.length - str.length; i++) {
                this.writeByte(0);
            }
        }
    }
    writeByteArray(list, options) {
        this.writeByte(list ? list.length : 0, options);
        (list || []).forEach((l) => {
            this.writeByte(l, options);
        });
    }
    writeData(type, val, options) {
        switch (type) {
            case DataTypes.BYTE:
                this.writeByte(val, options);
                break;
            case DataTypes.BOOL:
                this.writeBool(val, options);
                break;
            case DataTypes.INT8:
                this.writeByte(val, options);
                break;
            case DataTypes.INT16:
                this.writeInt16(val, options);
                break;
            case DataTypes.INT32:
                this.writeInt32(val, options);
                break;
            case DataTypes.UINT8:
                this.writeByte(val, options);
                break;
            case DataTypes.UINT16:
                this.writeUInt16(val, options);
                break;
            case DataTypes.UINT32:
                this.writeUInt32(val, options);
                break;
            case DataTypes.FLOAT:
                this.writeFloat(val, options);
                break;
            case DataTypes.DOUBLE:
                this.writeDouble(val, options);
                break;
            case DataTypes.DATE:
                this.writeDate(val, options);
                break;
            case DataTypes.STRING:
                this.writeString(val, options);
                break;
            case DataTypes.BYTE_ARRAY:
                this.writeByteArray(val, options);
                break;
        }
    }
    trimLeft(count) {
        const bytes = this._buffer.slice(0, count);
        this._buffer = Buffer.from(this._buffer.slice(count, this._buffer.length));
        this.index = this.index - count;
        if (this.index < 0) {
            this.index = 0;
        }
        this._length -= count;
        if (this._length < 0) {
            this._length = 0;
        }
        return bytes;
    }
    trimRight(count) {
        this.index -= count;
        if (this.index < 0) {
            this.index = 0;
        }
        this._length -= count;
        if (this._length < 0) {
            this._length = 0;
        }
    }
    readByte() {
        if (this.length < this.index + 1) {
            throw new Error('Buffer Overrun');
        }
        const byte = this._buffer.readUInt8(this.index);
        this.index++;
        return byte;
    }
    readInt8() {
        if (this.length < this.index + 1) {
            throw new Error('Buffer Overrun');
        }
        const byte = this._buffer.readInt8(this.index);
        this.index++;
        return byte;
    }
    readBool() {
        return this.readByte() === 1;
    }
    readInt16() {
        if (this.length < this.index + 2) {
            throw new Error('Buffer Overrun');
        }
        const val = this.useLittleEndian ? this._buffer.readInt16LE(this.index) : this._buffer.readInt16BE(this.index);
        this.index += 2;
        return val;
    }
    readInt32() {
        if (this.length < this.index + 4) {
            throw new Error('Buffer Overrun');
        }
        const val = this.useLittleEndian ? this._buffer.readInt32LE(this.index) : this._buffer.readInt32BE(this.index);
        this.index += 4;
        return val;
    }
    readUInt16() {
        if (this.length < this.index + 2) {
            throw new Error('Buffer Overrun');
        }
        const val = this.useLittleEndian ? this._buffer.readUInt16LE(this.index) : this._buffer.readUInt16BE(this.index);
        this.index += 2;
        return val;
    }
    readUInt32() {
        if (this.length < this.index + 4) {
            throw new Error('Buffer Overrun');
        }
        const val = this.useLittleEndian ? this._buffer.readUInt32LE(this.index) : this._buffer.readUInt32BE(this.index);
        this.index += 4;
        return val;
    }
    readFloat() {
        if (this.length < this.index + 8) {
            throw new Error('Buffer Overrun');
        }
        const val = this.useLittleEndian ? this._buffer.readFloatLE(this.index) : this._buffer.readFloatBE(this.index);
        this.index += 8;
        return val;
    }
    readDouble() {
        if (this.length < this.index + 8) {
            throw new Error('Buffer Overrun');
        }
        const val = this.useLittleEndian ? this._buffer.readDoubleLE(this.index) : this._buffer.readDoubleBE(this.index);
        this.index += 8;
        return val;
    }
    readDate() {
        try {
            const year = this.readByte() + 2000;
            const month = this.readByte();
            const day = this.readByte();
            const hour = this.readByte();
            const minute = this.readByte();
            const second = this.readByte();
            return new Date(Date.UTC(year, month, day, hour, minute, second));
        }
        catch (e) {
            return null;
        }
    }
    readBytes(count) {
        if (this.length < this.index + count) {
            throw new Error('Buffer Overrun');
        }
        const bytes = this._buffer.slice(this.index, this.index + count);
        this.index += count;
        return bytes;
    }
    readString(options = {}) {
        if (this.length < this.index + 2) {
            throw new Error('Buffer Overrun');
        }
        const length = options.length || this.readUInt16();
        if (this.length < this.index + length) {
            throw new Error('Buffer Overrun');
        }
        const str = this._buffer.toString('ascii', this.index, this.index + length).replace(/\0/g, '');
        this.index += length;
        return str;
    }
    readByteArray() {
        const array = [];
        const length = this.readByte();
        for (let i = 0; i < length; i++) {
            array.push(this.readByte());
        }
        return array;
    }
    readData(type, options = {}) {
        switch (type) {
            case DataTypes.BYTE:
                return this.readByte();
            case DataTypes.BOOL:
                return this.readBool();
            case DataTypes.INT8:
                return this.readInt8();
            case DataTypes.INT16:
                return this.readInt16();
            case DataTypes.INT32:
                return this.readInt32();
            case DataTypes.UINT8:
                return this.readByte();
            case DataTypes.UINT16:
                return this.readUInt16();
            case DataTypes.UINT32:
                return this.readUInt32();
            case DataTypes.FLOAT:
                return this.readFloat();
            case DataTypes.DOUBLE:
                return this.readDouble();
            case DataTypes.DATE:
                return this.readDate();
            case DataTypes.STRING:
                return this.readString(options);
            case DataTypes.BYTE_ARRAY:
                return this.readByteArray();
        }
    }
    toString() {
        let str = '';
        for (const byte of this._buffer) {
            str += byte.toString(16).toUpperCase() + ' ';
        }
        return str;
    }
    prepareBuffer(length) {
        const spaceLeft = this._buffer.length - this.length;
        if (length > spaceLeft) {
            this._buffer = Buffer.concat([this._buffer, Buffer.alloc(this._paddingSize + length)], this._buffer.length + this._paddingSize + length);
        }
    }
}
